import bisect
import random
import numpy as np
from itertools import product

MaxArcCount = 2  # 16;
ChainLength = 2
IterCountForBuild = 1000000  # 6000000 10000000 15000000;
StartPhero = 0.1
AntPhero = 0.1
AntiPhero = 0.005
AntStepCount = 3

IntervalCount = 40
IntervalStep = 0.1

SinusStartPoint = 0.0
SinusStep = 0.1  # Math.PI / 12.0; == 0.26
SinusNoise = 0.001
SinusElemCount = 24000

Y = 0.8
X = -1.0
Z = 0.2
LorentzElemCount = 50000
DELTA = 10.0
B = 8.0 / 3.0
R = 28
STEP = 0.1

InterCountForRebuild = 10
ChainCountForReBuild = 15


def normalize(series):
    '''
    
    normalize series according to 
    missing values (np.nan is missing values)
    
    input - time series 
    output - (normalized time series, (min values, max values))
    
    '''
    series_mins = np.nanmin(series, axis=0)
    series_maxes = np.nanmax(series, axis=0)
    series = 2.0 * (series - (series_mins + series_maxes) / 2) / (series_maxes - series_mins)
    return series, np.array([series_mins, series_maxes])


def get_intervals(series, interval_number=IntervalCount):
    '''
    
    get interval Ii time series according to number of partitions
    
    input - series, number of intervals
    output - interval for time series
    
    '''
    intervals = np.zeros((series.shape[1], interval_number - 1))
    for dim in range(series.shape[1]):
        v = 0
        sor = sorted(series[~np.isnan(series).any(axis=1)], key=lambda x: x[dim])
        d_interval = np.zeros(interval_number - 1)

        for i in range(interval_number - 1):
            v += int(1 / interval_number * len(sor))
            d_interval[i] = sor[v - 1][dim]

        intervals[dim, :] = d_interval
    return intervals


def sequence_generator(length, max_step):
    '''generate the pattern p (p1,...,ps-1) from article;
    length is s-1 from article; 
    max step in Rmax from article
    
    input s-1, Rmax
    output list like (p1,...,ps-1)
    '''
    return list(product(range(1, max_step + 1), repeat=length))


def interval_choose(value, interval):
    
    '''sort intervals according to value
    (sort the list after each insertion)
    https://docs.python.org/3/library/bisect.html
    '''
    return bisect.bisect_left(interval, value)


def convert_series_to_indices(series, intervals):
    '''convert series to indices
    input: series, interval Ii
    output: series of index of interval (I1...Is-1), average interval values
    
    '''
    interval_counts = np.zeros((intervals.shape[0], intervals.shape[1] + 1))
    interval_sum = np.zeros((intervals.shape[0], intervals.shape[1] + 1))
    average_interval_values = np.zeros((intervals.shape[0], intervals.shape[1] + 1))
    for i in range(series.shape[0]):
        for j in range(series.shape[1]):
            if not np.isnan(series[i][j]):
                interval_number = interval_choose(series[i][j], intervals[j])
                interval_counts[j][interval_number] += 1
                interval_sum[j][interval_number] += series[i][j]
                series[i][j] = interval_number
    average_interval_values = interval_sum / interval_number
    return series, average_interval_values


def append_in_tensor(relation_tensor, current_sequence, current_steps, current_pheromone, additional_pheromone):
    '''update tensor by ant alg. '''
    key = str([current_sequence, current_steps])
    if key not in relation_tensor:
        relation_tensor[key] = current_pheromone
    relation_tensor[key] += additional_pheromone


def make_steps(current_position, steps, series):
    '''step of ant 
    input current position, steps, series
    output Ii...'''
    intervals = [series[current_position]]
    for step in steps:
        current_position += step
        intervals.append(series[current_position])
    return intervals

def select_start_sequence(series, start_position, relation_tensor, current_pheromone_level, sequences_):
    '''An ant is placed on the tensor element defined by the initial conditions (for the first ant) or
       by already filled positions (for all other ants). In the latter case, from a set of already filled
       observations, the algorithm selects the one that immediately precedes the first unfilled one.
       It becomes the next start position.
       
       input series, start position (p1,...,ps-2), T(paper), current pheromone, sequences (ant way (paper))
       
       '''
    value_sum = 0
    probabilities = np.zeros(len(sequences_))
    for i in range(len(probabilities)):
        current_sequence_intervals = make_steps(start_position, sequences_[i], series)
        relation = str([current_sequence_intervals, sequences_[i]])
        if relation in relation_tensor:
            value_sum += relation_tensor[relation]
            probabilities[i] = relation_tensor[relation]
        elif relation.find('nan') == -1:
            value_sum += current_pheromone_level
            probabilities[i] = current_pheromone_level

    if value_sum == 0:
        return -1, 0
    probabilities /= value_sum
    '''Both position to which an ant transits and the interval to which the value of respective
      observation belongs are determined probabilistically, using the roulette wheel, with probabilities
      determined as'''
    sequence_steps = sequences_[np.random.choice(np.arange(len(probabilities)), p=probabilities)]
    return make_steps(start_position,sequence_steps, series), list(sequence_steps)


def select_next_step(series, current_position, current_sequence, current_steps, relation_tensor, current_pheromone_level,
                     max_step = MaxArcCount):
    '''simmilar to select_start_sequence
       2, 3 points in alg 2 in paper'''
    value_sum = 0
    probabilities = np.zeros(max_step)
    for step in range(1, max_step + 1):
        if np.isnan(series[current_position + step]).any() :
            continue

        relation = str([current_sequence + [series[current_position + step]], current_steps + [step]])
        if relation in relation_tensor:
            value_sum += relation_tensor[relation]
            probabilities[step - 1] = relation_tensor[relation]

        else:
            value_sum += current_pheromone_level
            probabilities[step - 1] = current_pheromone_level

    if value_sum == 0:
        return -1
    probabilities /= value_sum
    '''Both position to which an ant transits and the interval to which the value of respective
      observation belongs are determined probabilistically, using the roulette wheel, with probabilities
      determined as'''
    step = np.random.choice(np.arange(1, max_step + 1), p=probabilities)
    return step



def generate_relation_tensor(series, max_step=MaxArcCount, vector_len=ChainLength, interval_number=IntervalCount,
                             max_step_number=100, initial_pheromone=1, pheromone_transition_number=ChainLength,
                             additional_pheromone=1, max_transition_number=25):  # для многомерности добавить dimensions
    '''alg 1 from paper'''
    series, recovering_values = normalize(series)
    sequences_ = sequence_generator(vector_len, max_step)
    intervals = get_intervals(series, interval_number)
    series, average_interval_values = convert_series_to_indices(series, intervals)
    relation_tensor = {}
    current_pheromone_level = initial_pheromone
    step_counter = 0

    while step_counter < max_step_number:
        if not step_counter % 10000:
            print(step_counter // 10000, '% done', sep='')
        step_counter += 1
        start_position = random.randrange(len(series) - vector_len * max_step)
        while np.isnan(series[start_position]).any():
            start_position = random.randrange(len(series) - vector_len * max_step)

        current_sequence, current_steps = select_start_sequence(series, start_position, relation_tensor, current_pheromone_level,
                                                 sequences_)
        if current_sequence == -1:
            continue

        current_position = start_position + sum(current_steps)
        transition_counter = vector_len

        while current_position + max_step < len(series) and transition_counter < max_transition_number:
            transition_counter += 1
            current_step = select_next_step(series, current_position, current_sequence, current_steps, relation_tensor,
                                            current_pheromone_level, max_step)
            if current_step == -1:
                break
            current_position = current_position + current_step
            current_sequence= np.append(current_sequence[1:],[series[current_position]], axis= 0)
            current_steps = np.append(current_steps[1:],current_step)

            if not transition_counter % pheromone_transition_number:
                append_in_tensor(relation_tensor, current_sequence, current_steps,
                                 current_pheromone_level, additional_pheromone)

    for i in relation_tensor.keys():
        relation_tensor[i] -= initial_pheromone
    return relation_tensor, average_interval_values, recovering_values, intervals




def accurate_generate_relation_tensor(series, max_step=MaxArcCount, vector_len=ChainLength,
                                      interval_number=IntervalCount, additional_pheromone=1):  # для многомерности добавить dimensions
    series, recovering_values = normalize(series)
    sequences_ = sequence_generator(vector_len, max_step)
    intervals = get_intervals(series,interval_number)
    series,average_interval_values = convert_series_to_indices(series, intervals)
    relation_tensor = {}
    for j in sequences_:
        current_position = np.cumsum([0] + list(j))
        for i in range(len(series) - max_step * vector_len):
            current_sequence = series[current_position, :]
            if not np.isnan(current_sequence).any():
                append_in_tensor(relation_tensor, current_sequence, list(j), 0, additional_pheromone)
            current_position += 1
    return relation_tensor, average_interval_values, recovering_values
